﻿using System;
using System.IO;
using System.Linq;
using System.Numerics;
using System.Text;

using NAudio.Wave;

namespace DECODER;

/// <summary>
/// 디코더
/// </summary>
class Decoder
{
    //////////////////////////////////////////////////////////////////////////////////////////////////// Method
    ////////////////////////////////////////////////////////////////////////////////////////// Static
    //////////////////////////////////////////////////////////////////////////////// Private

    #region 다음 2의 거듭 제곱 구하기 - NextPowerOfTwo(value)

    /// <summary>
    /// 다음 2의 거듭 제곱 구하기
    /// </summary>
    /// <param name="value">값</param>
    /// <returns></returns>
    /// <remarks>입력값보다 큰 가장 작은 2의 거듭 제곱을 계산한다.</remarks>
    private static int NextPowerOfTwo(int value)
    {
        int power = 1;

        while(power < value)
        {
            power <<= 1;
        }

        return power;
    }

    #endregion
    #region 비트 반전하기 - BitReverse(bitValue, bitCount)

    /// <summary>
    /// 비트 반전하기
    /// </summary>
    /// <param name="bitValue">비트 값</param>
    /// <param name="bitCount">비트 수</param>
    /// <returns>반전 비트 값</returns>
    private static int BitReverse(int bitValue, int bitCount)
    {
        int reversed = 0;

        for(int i = 0; i < bitCount; i++)
        {
            int nextBit = (bitValue >> i) & 1;

            reversed |= (nextBit << (bitCount - i - 1));
        }

        return reversed;
    }

    #endregion
    #region FFT 처리하기 - FFT(complexArray)

    /// <summary>
    /// FFT 처리하기
    /// </summary>
    /// <param name="complexArray">복소수 배열</param>
    private static void FFT(Complex[] complexArray)
    {
        int complexArrayLength = complexArray.Length;

        if(complexArrayLength <= 1)
        {
            return;
        }

        // Bit reversal permutation
        int m1 = (int)Math.Log(complexArrayLength, 2);

        for(int i = 0; i < complexArrayLength; i++)
        {
            int j = BitReverse(i, m1);

            if(j > i)
            {
                Complex t = complexArray[i];

                complexArray[i] = complexArray[j];
                complexArray[j] = t;
            }
        }

        // Cooley-Tukey FFT
        for(int s = 1; s <= m1; s++)
        {
            int m2 = 1 << s;

            Complex wm = Complex.Exp(-2.0 * Math.PI * Complex.ImaginaryOne / m2);

            for(int k = 0; k < complexArrayLength; k += m2)
            {
                Complex w = 1;

                for(int j = 0; j < m2 / 2; j++)
                {
                    int index1 = k + j;
                    int index2 = k + j + m2 / 2;

                    Complex t = w * complexArray[index2];
                    Complex u = complexArray[index1];

                    complexArray[index1] = u + t;
                    complexArray[index2] = u - t;

                    w *= wm;
                }
            }
        }
    }

    #endregion
    #region 프로그램 시작하기 - Main(argumentArray)

    /// <summary>
    /// 프로그램 시작하기
    /// </summary>
    /// <param name="argumentArray">인자 배열</param>
    private static void Main(string[] argumentArray)
    {
        if(argumentArray.Length != 2)
        {
            Console.WriteLine("사용법 : decoder.exe source.wav target.zip");

            return;
        }

        string sourceFilePath = argumentArray[0];
        string targetFilePath = argumentArray[1];

        int    samplingFrequency = 44100; // 샘플링 주파수
        double durationPerBit    = 0.01;  // 각 비트의 지속 시간 (초)
        int    sampleCountPerBit = (int)(samplingFrequency * durationPerBit);

        int frequency0 = 1000; // '0'에 대한 주파수
        int frequency1 = 2000; // '1'에 대한 주파수

        // 시작 및 종료 마커를 설정한다.
        string startMarkerBitString = "1111000011110000";
        string endMarkerBitString   = "0000111100001111";

        // WAV 파일을 읽는다.
        float[] signalArray;

        using(AudioFileReader audioFileReader = new(sourceFilePath))
        {
            int totalSampleCount = (int)audioFileReader.Length / (audioFileReader.WaveFormat.BitsPerSample / 8);

            signalArray = new float[totalSampleCount];

            int readSampleCount = audioFileReader.Read(signalArray, 0, totalSampleCount);
        }

        int bitCount = signalArray.Length / sampleCountPerBit;

        StringBuilder bitStringBuilder = new StringBuilder();

        for(int i = 0; i < bitCount; i++)
        {
            int start = i * sampleCountPerBit;
            int end   = start + sampleCountPerBit;

            if(end > signalArray.Length)
            {
                break;
            }

            float[] segmentArray = new float[sampleCountPerBit];

            Array.Copy(signalArray, start, segmentArray, 0, sampleCountPerBit);

            // 입력 배열 길이를 가장 가까운 2의 거듭제곱으로 패딩한다.
            int paddedLength = NextPowerOfTwo(segmentArray.Length);

            Complex[] fftComplexArray = new Complex[paddedLength];

            for(int j = 0; j < segmentArray.Length; j++)
            {
                fftComplexArray[j] = new Complex(segmentArray[j], 0);
            }

            for(int j = segmentArray.Length; j < paddedLength; j++)
            {
                fftComplexArray[j] = Complex.Zero;
            }

            // 주파수 분석을 위한 FFT를 수행한다.
            FFT(fftComplexArray);

            double[] magnitudeArray = fftComplexArray.Take(paddedLength / 2).Select(c => c.Magnitude).ToArray();

            int peakIndex = Array.IndexOf(magnitudeArray, magnitudeArray.Max());

            double peakFrequency = (double)peakIndex * samplingFrequency / paddedLength;

            // 주파수에 따라 비트를 결정한다.
            if(Math.Abs(peakFrequency - frequency0) < 100)
            {
                bitStringBuilder.Append('0');
            }
            else if(Math.Abs(peakFrequency - frequency1) < 100)
            {
                bitStringBuilder.Append('1');
            }
            else
            {
                bitStringBuilder.Append('?');
            }
        }

        string bitStream = bitStringBuilder.ToString();

        // 비트 스트림에서 시작 및 종료 마커를 찾는다.
        int startMarkerIndex = bitStream.IndexOf(startMarkerBitString);
        int endMarkerIndex   = bitStream.IndexOf(endMarkerBitString);

        if(startMarkerIndex == -1 || endMarkerIndex == -1)
        {
            Console.WriteLine("시작 또는 종료 마커를 찾을 수 없습니다.");

            return;
        }

        int dataStart = startMarkerIndex + startMarkerBitString.Length;
        int dataEnd   = endMarkerIndex;

        string dataBitStream = bitStream.Substring(dataStart, dataEnd - dataStart);

        // 파일 크기를 추출한다. (첫 32비트)
        if(dataBitStream.Length < 32)
        {
            Console.WriteLine("파일 크기 정보를 읽을 수 없습니다.");

            return;
        }

        string fileSizeString = dataBitStream.Substring(0, 32);

        dataBitStream = dataBitStream.Substring(32);

        // 파일 크기를 정수로 변환한다.
        int fileSize = Convert.ToInt32(fileSizeString, 2);

        // 결정할 수 없는 비트를 제거한다.
        dataBitStream = dataBitStream.Replace("?", "");

        // 비트를 바이트로 변환한다.
        int byteCount = dataBitStream.Length / 8;

        byte[] dataByteArray = new byte[byteCount];

        for(int i = 0; i < byteCount; i++)
        {
            string byteBitString = dataBitStream.Substring(i * 8, 8);

            dataByteArray[i] = Convert.ToByte(byteBitString, 2);
        }

        // 파일 크기에 따라 데이터를 자른다.
        if(dataByteArray.Length > fileSize)
        {
            byte[] tempByteArray = new byte[fileSize];

            Array.Copy(dataByteArray, tempByteArray, fileSize);

            dataByteArray = tempByteArray;
        }

        // 복원된 파일을 저장한다.
        File.WriteAllBytes(targetFilePath, dataByteArray);

        Console.WriteLine("디코딩 완료 : " + targetFilePath);
    }

    #endregion
}